#!/usr/bin/env python

###############################
# 20190508
#
#
# At some point this will be used to read the all the results for an event,
# and output the text file with the marg ll per event, and the text file with
# the extrinsic parameters and ll for the loudest x intrinsic grid points.
#
# For now It jus reads in margll.txt, takes the loudest file, reads the
# samples file for that, and outputs it to a text file
#
# Commands are:
# convert_result_to_txt.py <path_to_event_dir>
#
#############################

__author__ = "Sinead Walsh, Vinaya Valsan"


import sys
import glob
import os

import numpy as np

from argparse import ArgumentParser
from ligo.lw import utils, lsctables, ligolw
from rapid_pe import amrlib

module_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(module_dir)

lsctables.use_in(ligolw.LIGOLWContentHandler)

optp = ArgumentParser()
optp.add_argument("input_dir", help="path to event run dir")
optp.add_argument( "--ratio-to-include", type=float, default=None, help="consider only the grid points with margll>=ratio_to_include*max_margll")
optp.add_argument( "--level", default="", help="if specified considers results file from a specific level")
optp.add_argument( "--output-dir", default=None, help="directory to save plots")
opts = optp.parse_args()

print("-----------------Converting results to txt--------------------")
input_dir = opts.input_dir

ratio_to_include =  opts.ratio_to_include

if opts.output_dir:
    output_dir = opts.output_dir
else:
    output_dir = input_dir

level = str(opts.level)
print("LEVEL", level)

results_dir = os.path.join(input_dir,"results")

margll_summary_filename = output_dir + "/margll" + level + ".txt"
# in v2, we exclude candidates with very low weights
ll_loudest_filename = (
    output_dir
    + "/ll_samples_loudest_"
    + str(ratio_to_include).replace(".", "p")
    + level
    + ".txt"
)
ll_loudest_filename_highweight = (
    output_dir
    + "/ll_samples_loudest_highweight_"
    + str(ratio_to_include).replace(".", "p")
    + level
    + ".txt"
)
read_log_info = 0
log_info_filename = output_dir + "/log_files_summary" + level + ".txt"
force_reread = 0


def main():

    """
    Read rapidPE results. output to txt file.
    Separate text file for
    - margll per intrinsic point
    - ll per extrinsic points for loudest intrinsic points
    - info from log files e.g. neff
    """

    input_files = []
    if level == "":
        input_files = glob.glob(results_dir + "/ILE_iteration_*.xml.gz")
    else:
        print("WARNING: loading only level", level)
        input_files = glob.glob(
            results_dir + "/ILE_iteration_" + level + "*.xml.gz"
        )
    print(len(input_files), "files in results directory.", results_dir)

    # Check if the margll file contains all results, if not, reread margll data
    margll = {}
    reread_margll = 1
    if os.path.isfile(margll_summary_filename) and not force_reread:
        # if the margll from each file has already been written to a text
        # file, read the text file
        margll = np.atleast_1d(
            np.genfromtxt(
                margll_summary_filename,
                names=True,
                dtype=None,
                encoding=None,
            )
        )

        if len(input_files) == len(margll["margll"]) * 2:
            reread_margll = 0

    # Read margll for each intrinsic point
    if reread_margll:
        print("REREADING MARGLL")
        margll_output_lines = (
            "mass1 mass2 mchirp eta spin1z spin2z margll filename mass_set\n"
        )
        for filename in input_files:
            if ".samples.xml.gz" in filename:
                continue

            print(filename)
            xmldoc = utils.load_filename(
                filename, contenthandler=ligolw.LIGOLWContentHandler
            )
            # FIXME: can I tell from lsctables.SnglInspiralTable.get_table
            # if it's samples or not?
            new_tbl = lsctables.SnglInspiralTable.get_table(xmldoc)

            # Transform input mass parameters to parameters I want to plot
            for row in new_tbl:
                row.mchirp, row.eta = amrlib.transform_m1m2_mceta(
                    row.mass1, row.mass2
                )
                try:
                    margll_output_lines += (
                        f"{str(row.mass1)} {str(row.mass2)} {str(row.mchirp)} "
                        f"{str(row.eta)} {str(row.spin1z)} {str(row.spin2z)} "
                        f"{str(row.snr)} {filename} "
                    )
                except:
                    margll_output_lines += (
                        f"{str(row.mass1)} {str(row.mass2)} {str(row.mchirp)} "
                        f"{str(row.eta)} 0.0 0.0 "
                        f"{str(row.snr)} {filename} "
                    )
                msi1 = filename[filename.rfind("MASS_SET_") + 9:]
                mass_set_id = msi1[: msi1.find("-")]
                margll_output_lines += mass_set_id + "\n"

            xmldoc.unlink()

        fo = open(margll_summary_filename, "w")
        fo.write(margll_output_lines)
        fo.close()
        print(
            "Intrinsic params margll summary written to ",
            margll_summary_filename,
        )

    # now read it back in to find the loudest filenaems
    margll = np.atleast_1d(
        np.genfromtxt(
            margll_summary_filename,
            names=True,
            dtype=None,
            encoding=None,
        )
    )

    if (
        os.path.isfile(ll_loudest_filename)
        and not reread_margll
        and not force_reread
    ):
        sys.exit("Skipping file already read")

    # Extrinsic params
    # Figure out which indices belong to the intrinsic points wiht the
    # highest likelihood
    # Read the samples files for these and save the likelihoods calculated
    # for the extrinsic parameters
    max_margll = max(margll["margll"])
    loudest_indices = (
        np.where(margll["margll"] >= ratio_to_include * max_margll)[0]
        if len(margll["margll"]) > 1
        else [0]
    )
    #    loudest_indices = np.where(margll["margll"] >= 0.01)[0]
    loudest_filenames = margll["filename"][loudest_indices]
    samples_output_lines = (
        "mass1 mass2 mchirp eta spin1z spin2z distance latitude longitude "
        "inclination phase polarization likelihood prior sampling_function\n"
    )
    print("Loudest files", len(loudest_filenames), "\n", loudest_filenames)
    all_lls = []
    for filename in loudest_filenames:
        try:
            samples_filename = filename  # .replace(".xml.gz", ".samples.xml.gz")
            xmldoc = utils.load_filename(
                samples_filename, contenthandler=ligolw.LIGOLWContentHandler
            )
            new_tbl = lsctables.SimInspiralTable.get_table(xmldoc)
        except:
            samples_filename = filename.replace(".xml.gz", ".samples.xml.gz")
            xmldoc = utils.load_filename(
                samples_filename, contenthandler=ligolw.LIGOLWContentHandler
            )
            new_tbl = lsctables.SimInspiralTable.get_table(xmldoc)

        # Transform input mass parameters to parameters I want to plot
        for row in new_tbl:
            row.mchirp, row.eta = amrlib.transform_m1m2_mceta(
                row.mass1, row.mass2
            )

            all_lls.append(row.alpha1)
            samples_output_lines += (
                f"{str(row.mass1)} {str(row.mass2)} "
                f"{str(row.mchirp)} {str(row.eta)} "
                f"{str(row.spin1z)} {str(row.spin2z)} "
                f"{str(row.distance)} {str(row.latitude)} "
                f"{str(row.longitude)} {str(row.inclination)} "
                f"{str(row.coa_phase)} {str(row.polarization)} "
                f"{str(row.alpha1)} {str(row.alpha2)} {str(row.alpha3)}\n"
            )

    #    print ("Extrinsic params ll for loudest margll summary written
    #    to ",ll_loudest_filename)
    print(
        "Extrinsic params ll for loudest margll summary No longer written to ",
        ll_loudest_filename,
        "to avoid excessively large files, only highweight written",
    )
    #    fo = open(ll_loudest_filename,"w")
    #    fo.write(samples_output_lines)
    #    fo.close()
    #    os.system("wc -l "+ll_loudest_filename)

    # reread everything. This time, exclude those with low weights.
    maxlnl = max(all_lls)
    samples_output_lines = (
        "mass1 mass2 mchirp eta spin1z spin2z distance latitude longitude "
        "inclination phase polarization likelihood prior sampling_function "
        "weight\n"
    )
    for filename in loudest_filenames:
        try:
            samples_filename = filename  # .replace(".xml.gz", ".samples.xml.gz")

            xmldoc = utils.load_filename(
                samples_filename, contenthandler=ligolw.LIGOLWContentHandler
            )
            new_tbl = lsctables.SimInspiralTable.get_table(xmldoc)
        except:
            samples_filename = filename.replace(".xml.gz", ".samples.xml.gz")

            xmldoc = utils.load_filename(
                samples_filename, contenthandler=ligolw.LIGOLWContentHandler
            )
            new_tbl = lsctables.SimInspiralTable.get_table(xmldoc)

        # Transform input mass parameters to parameters I want to plot
        for row in new_tbl:
            row.mchirp, row.eta = amrlib.transform_m1m2_mceta(
                row.mass1, row.mass2
            )

            # alpha1=log likelihood. alpha2 = prior. alpha3 = sampling_function
            weight = (np.exp(row.alpha1 - maxlnl) * row.alpha2) / row.alpha3
            if weight < 1e-7:
                continue

            samples_output_lines += (
                f"{str(row.mass1)} {str(row.mass2)} "
                f"{str(row.mchirp)} {str(row.eta)} "
                f"{str(row.spin1z)} {str(row.spin2z)} "
                f"{str(row.distance)} {str(row.latitude)} "
                f"{str(row.longitude)} {str(row.inclination)} "
                f"{str(row.coa_phase)} {str(row.polarization)} "
                f"{str(row.alpha1)} {str(row.alpha2)} "
                f"{str(row.alpha3)} {str(weight)}\n"
            )
    print(
        f"Extrinsic params ll for loudest margll summary, excluding low "
        f"weight samples, written to {ll_loudest_filename_highweight}"
    )
    fo = open(ll_loudest_filename_highweight, "w")
    fo.write(samples_output_lines)
    fo.close()
    os.system("wc -l " + ll_loudest_filename_highweight)

    # Read in the log information, and save it
    if read_log_info and (
        reread_margll or force_reread or not os.path.isfile(log_info_filename)
    ):
        log_summary_lines = "neff per_point_time mass_set\n"
        for filename in input_files:
            if ".samples.xml.gz" in filename:
                continue

            # Read in the corresponding log file
            out_file_start = filename[: filename.rfind("results/")]
            out_file_end = filename[filename.rfind("ILE_iteration_") + 15: -6]
            out_file = out_file_start + "logs/integrate" + out_file_end + "out"
            outlog = open(out_file, "r")
            start_time = end_time = -1
            for line in outlog:
                if "walltime: iteration Neff" in line:
                    start_time = 0
                    continue
                if start_time == 0:
                    start_time = float(line.split(" ")[0])
                    continue
                if (
                    start_time > 0
                    and not "lnLmarg is" in line
                    and not "note neff" in line
                ):
                    end_time = float(line.split(" ")[0])
                    continue
                if "note neff is" in line:
                    tmp1 = line.replace("note neff is", "")
                    tmp2 = tmp1.replace("\n", "")
                    log_summary_lines += str(tmp2.replace(" ", ""))
            log_summary_lines += " " + str((end_time - start_time) / 60.0)

            msi1 = filename[filename.rfind("MASS_SET_") + 9:]
            mass_set_id = msi1[: msi1.find("-")]
            log_summary_lines += " " + mass_set_id + "\n"

        print("Log information written to ", log_info_filename)
        fo = open(log_info_filename, "w")
        fo.write(log_summary_lines)
        fo.close()

    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

__author__ = "Anarya Ray, Caitlin Rose, Vinaya Valsan"


import os
import scipy

import numpy as np
import matplotlib.pyplot as plt

from gstlal import far


class pastro:
    def __init__(
        self,
        evidence_dict,
        rankstatpdf_file,
        likelihood,
        rate_dict,
        far_threshold,
    ):
        self.evidence_dict = evidence_dict
        self.rate_dict = rate_dict
        self.far_threshold = far_threshold
        self.rankstatpdf_file = rankstatpdf_file
        self.likelihood = likelihood

    def normalize_f_over_b(self):
        rankingstatpdf = far.marginalize_pdf_urls(
            [self.rankstatpdf_file], "RankingStatPDF"
        )

        zl = rankingstatpdf.zero_lag_lr_lnpdf.copy()
        zl.array[:40] = 0.0

        if not zl.array.any():
            raise ValueError("zero-lag counts are all zero")
        (ln_likelihood_ratio_threshold,) = zl.argmax()

        rankingstatpdf = rankingstatpdf.new_with_extinction()
        fapfar = far.FAPFAR(rankingstatpdf)

        xmin = max(
            fapfar.rank_from_far(self.far_threshold),
            ln_likelihood_ratio_threshold,
        )

        rankingstatpdf.noise_lr_lnpdf.array[
            : rankingstatpdf.noise_lr_lnpdf.bins[0][xmin]
        ] = 0.0
        rankingstatpdf.noise_lr_lnpdf.normalize()
        rankingstatpdf.signal_lr_lnpdf.array[
            : rankingstatpdf.signal_lr_lnpdf.bins[0][xmin]
        ] = 0.0
        rankingstatpdf.signal_lr_lnpdf.normalize()
        rankingstatpdf.zero_lag_lr_lnpdf.array[
            : rankingstatpdf.zero_lag_lr_lnpdf.bins[0][xmin]
        ] = 0.0
        rankingstatpdf.zero_lag_lr_lnpdf.normalize()

        noise_lr_lnpdf = rankingstatpdf.noise_lr_lnpdf.mkinterp()
        lnb = np.vectorize(lambda x: noise_lr_lnpdf(x))

        B, dB = scipy.integrate.quad(
            lambda x: np.exp(lnb(x)), xmin, np.inf, limit=1000
        )

        A, dA = scipy.integrate.quad(
            lambda x: np.exp(x + lnb(x) - np.log(B)), xmin, np.inf, limit=1000
        )

        return np.reciprocal(A)

    def pipeline_bayesfactor(self):
        return self.normalize_f_over_b() * np.exp(self.likelihood)

    def compute_pastro(self):
        R_tot = sum(self.rate_dict.values())
        print(f"pipeline_bayesfactor = {self.pipeline_bayesfactor()}")

        evidence_sum = (
            1.0
            if self.pipeline_bayesfactor() == 1.0
            else (
                self.evidence_dict["evidence_BNS"]
                + self.evidence_dict["evidence_NSBH"]
                + self.evidence_dict["evidence_BBH"]
            )
            / 3.0
        )
        pH = {
            "Terrestrial": self.far_threshold / R_tot,
            "BBH": self.rate_dict["R_bbh"] / R_tot,
            "NSBH": self.rate_dict["R_nsbh"] / R_tot,
            "BNS": self.rate_dict["R_bns"] / R_tot,
        }

        evidence_factor = self.pipeline_bayesfactor() / evidence_sum

        evidence_ratio = {
            "Terrestrial": 1.0,
            "BBH": self.evidence_dict["evidence_BBH"] * evidence_factor,
            "NSBH": self.evidence_dict["evidence_NSBH"] * evidence_factor,
            "BNS": self.evidence_dict["evidence_BNS"] * evidence_factor,
        }

        pastros_numerator = {k: z * pH[k] for k, z in evidence_ratio.items()}

        pastros_denominator = sum(pastros_numerator.values())
        pastros = {
            k: numerator / pastros_denominator
            for k, numerator in pastros_numerator.items()
        }
        return pastros


def plot_pastro(pastro_dict, save_dir, filename="p_astro.png"):
    """
    adapted from gwcelery.tasks
    """

    def _format_prob(prob):
        if prob >= 1:
            return "100%"
        elif prob <= 0:
            return "0%"
        elif prob > 0.99:
            return ">99%"
        elif prob < 0.01:
            return "<1%"
        else:
            return "{}%".format(int(np.round(100 * prob)))

    fname = os.path.join(save_dir, filename)

    probs = list(pastro_dict.values())
    names = list(pastro_dict.keys())

    with plt.style.context("seaborn-white"):
        fig, ax = plt.subplots(figsize=(2.5, 2))
        ax.barh(names, probs)
        for i, prob in enumerate(probs):
            ax.annotate(
                _format_prob(prob),
                (0, i),
                (4, 0),
                textcoords="offset points",
                ha="left",
                va="center",
            )
        ax.set_xlim(0, 1)
        ax.set_xticks([])
        ax.tick_params(left=False)
        for side in ["top", "bottom", "right"]:
            ax.spines[side].set_visible(False)
        fig.tight_layout()
        fig.savefig(fname)
    return



#!/usr/bin/env python3

__author__ = "Soichiro Morisaki, Vinaya Valsan"

import numpy as np
from rapid_pe import amrlib


# Jacobian used in posterior calculations.


def uniform_m1_m2_prior_in_mchirp_eta(mchirp, eta):
    """
    Returns  jacobian  p(mchirp, eta) = d(mass1,mass2)/d(mchirp,eta)
    """
    p = np.abs(mchirp * eta ** (-1.2) * (1.0 - 4.0 * eta) ** (-0.5))
    return p


def uniform_m1_m2_prior_in_mchirp_q(mchirp, q):
    """
    Returns  jacobian  p(mchirp, q) = d(mass1,mass2)/d(mchirp,q)
    """
    p = np.abs(mchirp * (1.0 + q) ** 0.4 * q ** (-1.2))
    return p


def uniform_m1_m2_prior_in_mtotal_q(mtotal, q):
    """
    Returns  jacobian  p(mtotal, q) = d(mass1,mass2)/d(mtotal,q)
    """
    p = np.abs(mtotal * (1 + q) ** (-2))
    return p


def uniform_m1_m2_prior_in_tau0_tau3(tau0, tau3, f_lower):
    """
    Returns  jacobian  p(tau0, tau3) = d(mass1,mass2)/d(tau0,tau3)
    """
    a3 = np.pi / (8.0 * (np.pi * f_lower) ** (5.0 / 3.0))
    a0 = 5.0 / (256.0 * (np.pi * f_lower) ** (8.0 / 3.0))
    tmp1 = (a0 * tau3) / (a3 * tau0)
    num = a0 * (tmp1) ** (1.0 / 3.0)
    tmp2 = 1 - ((4 * a3) / (tau3 * tmp1 ** (2.0 / 3.0)))
    den = tau0**2.0 * tau3 * np.sqrt(tmp2)
    return np.abs(num / den)


def uniform_m1m2chi1chi2_prior_to_mu1mu2qchi2(mu1, mu2, q, s2z):
    """Return d(mu1, mu2, q, s2z) / d(m1, m2, s1z, s2z)"""
    MsunToTime = 4.92659 * 10.0 ** (
        -6.0
    )  # conversion from solar mass to seconds
    fref_mu = 200.0
    # coefficients of mu1 and mu2
    mu_coeffs = np.array(
        [
            [0.97437198, 0.20868103, 0.08397302],
            [-0.22132704, 0.82273827, 0.52356096],
        ]
    )
    m1, m2, s1z, s2z = amrlib.transform_mu1mu2qs2z_m1m2s1zs2z(mu1, mu2, q, s2z)
    mc = (m1 * m2) ** (3.0 / 5.0) / (m1 + m2) ** (1.0 / 5.0)
    q = m2 / m1
    eta = amrlib.qToeta(q)
    x = np.pi * mc * MsunToTime * fref_mu
    tmp1 = (
        mu_coeffs[0, 2] * mu_coeffs[1, 0] - mu_coeffs[0, 0] * mu_coeffs[1, 2]
    )
    tmp2 = (
        mu_coeffs[0, 2] * mu_coeffs[1, 1] - mu_coeffs[0, 1] * mu_coeffs[1, 2]
    )
    denominator = (
        x
        * 5.0
        * (113.0 + 75.0 * q)
        * (
            252.0 * tmp1 * q * eta ** (-3.0 / 5.0)
            + tmp2 * (743.0 + 2410.0 * q + 743.0 * q**2.0) * x ** (2.0 / 3.0)
        )
    )
    numerator = (
        m1**2.0 * 4128768.0 * q * (1.0 + q) ** 2.0 * x ** (10.0 / 3.0)
    )
    return np.abs(numerator / denominator)


PRIOR_MAP = {
    "mchirp_eta": uniform_m1_m2_prior_in_mchirp_eta,
    "tau0_tau3": uniform_m1_m2_prior_in_tau0_tau3,
    "mchirp_q": uniform_m1_m2_prior_in_mchirp_q,
    "mtotal_q": uniform_m1_m2_prior_in_mtotal_q,
    "mu1_mu2_q_s2q": uniform_m1m2chi1chi2_prior_to_mu1mu2qchi2,
}
